import numpy as np
import pandas as pd
from anndata import AnnData
import warnings

def _check_adata(adata, column):
    if not isinstance(adata, AnnData):
        warnings.warn('adata does not appear to be an anndata object', UserWarning)
    if 'qc_constraints' not in adata.uns.keys():
        adata.uns['qc_constraints'] = {}
    if column not in adata.uns['qc_constraints'].keys():
        adata.uns['qc_constraints'][column] = {}

def add_range_constraint(adata: AnnData, 
                            column: str, 
                            gt: float = None, 
                            lt: float = None, 
                            subset: str = None, 
                            subset_values: list = None):
    """
    Adds a range constraint to a specific column in the AnnData object.

    This function filters cells where values in the specified column fall within the range 
    [gt, lt]. Subsetting based on another column is supported.

    Args:
        adata: The AnnData object where constraints will be added.
        column: The column in 'adata.obs' for which the range constraint will be applied.
        gt: The lower bound of the range (inclusive). If None, the lower bound is open.
        lt: The upper bound of the range (inclusive). If None, the upper bound is open.
        subset: The column in 'adata.obs' to use for subsetting the data.
        subset_values: A list of values for subsetting, applied to the 'subset' column.
    """
    constraint = {
        'gt': gt,
        'lt': lt,
        'subset': {},
        'exclude': [],
        'groupby': None
    }

    if subset and subset_values:
        constraint['subset'][subset] = subset_values

    _check_adata(adata, column)
    
    # Assign a unique key for each constraint
    constraint_key = f'constraint_{len(adata.uns["qc_constraints"][column]) + 1}'
    adata.uns['qc_constraints'][column][constraint_key] = constraint

def add_exclude_constraint(adata: AnnData, 
                            column: str, 
                            exclude_values: list, 
                            subset: str = None, 
                            subset_values: list = None):
    """
    Adds an exclude constraint to a specific column in the AnnData object.

    This function filters out cells where the value in the specified column is in 'exclude_values'.
    Subsetting based on another column is supported.

    Args:
        adata: The AnnData object where constraints will be added.
        column: The column in 'adata.obs' for which the exclude constraint will be applied.
        exclude_values: The list of values that should be excluded from 'column'.
        subset: The column in 'adata.obs' to use for subsetting the data.
        subset_values: A list of values for subsetting, applied to the 'subset' column.
    """
    
    constraint = {
        'gt': None,
        'lt': None,
        'subset': {},
        'exclude': exclude_values,
        'groupby': None
    }

    if subset and subset_values:
        constraint['subset'][subset] = subset_values

    _check_adata(adata, column)
    
    # Assign a unique key for each constraint
    constraint_key = f'constraint_{len(adata.uns["qc_constraints"][column]) + 1}'
    adata.uns['qc_constraints'][column][constraint_key] = constraint

def add_group_level_constraint(adata: AnnData, 
                                column: str, 
                                groupby: str, 
                                gt: float = None, 
                                lt: float = None, 
                                agg_func: str = 'mean'
                            ):
    """
    Adds a group-level constraint to a specific column in the AnnData object.

    Cells are grouped by 'groupby', and an aggregation function (e.g., 'mean') is applied to each group
    for the specified column. Cells in groups that meet the 'gt' and 'lt' conditions are kept.

    Args:
        adata: The AnnData object where constraints will be added.
        column: The column in 'adata.obs' for which the group-level constraint will be applied.
        groupby: The column in 'adata.obs' used to group the cells (e.g., cluster ID).
        gt: The lower bound for the aggregated group value. If None, the lower bound is open.
        lt: The upper bound for the aggregated group value. If None, the upper bound is open.
        agg_func: The aggregation function to apply to the groups. Can be 'mean', 'sum', 
                                  'std', or 'median'. Default is 'mean'.

    """
    constraint = {
        'gt': gt,
        'lt': lt,
        'subset': {},
        'exclude': [],
        'groupby': groupby,
        'agg_func': agg_func
    }

    _check_adata(adata, column)
    
    # Assign a unique key for each constraint
    constraint_key = f'constraint_{len(adata.uns["qc_constraints"][column]) + 1}'
    adata.uns['qc_constraints'][column][constraint_key] = constraint

def apply_constraints(adata: AnnData, 
                        inplace: bool = False
                    ) -> AnnData:
    """
    Applies all quality control (QC) constraints stored in 'adata.uns["qc_constraints"]' to 'adata.obs' and filtered indices in 'adata.uns["qc_filtered"]'.

    This function handles both individual cell-level constraints (range, exclude) and group-level constraints
    (using a 'groupby' column). The resulting 'keeper_cells' column in 'adata.obs' indicates which cells passed the constraints.

    Args:
        adata: The AnnData object where constraints are applied.
        inplace: If True, only the cells passing the constraints will be retained in 'adata'.
                 If False (default), the 'keeper_cells' column will be added to 'adata.obs' 
                 to indicate which cells passed the filtering.

    Returns:
        AnnData: The filtered AnnData object with the 'keeper_cells' column added to 'adata.obs'. If 'inplace=True', 
                 the returned object will only contain the filtered cells.
    """
    if isinstance(adata, AnnData):
        keeper_cells = pd.Series([True] * adata.n_obs, index=adata.obs.index)
        adata.uns['qc_filtered'] = {}
        agg_funcs = {'sum': np.sum, 'mean': np.mean, 'std': np.std, 'median': np.median}

        keys_to_remove = [] 
        
        for col, constraints_dict in adata.uns['qc_constraints'].items():
            if col not in adata.obs.columns:
                warnings.warn(f'Column {col} not in obs, skipping and removing this metric', UserWarning)
                keys_to_remove.append(col)
                continue

            column_data = adata.obs[col]
            col_keeper = pd.Series([True] * adata.n_obs, index=adata.obs.index)
            adata.uns['qc_filtered'][col] = {}
            
            # Iterate over the dictionary of constraints
            for constraint_key, constraints in constraints_dict.items():
                gt = pd.Series([True] * adata.n_obs, index=adata.obs.index)
                lt = pd.Series([True] * adata.n_obs, index=adata.obs.index)
                exclude = pd.Series([True] * adata.n_obs, index=adata.obs.index)
                subset = pd.Series([True] * adata.n_obs, index=adata.obs.index)

                # Apply group-level constraint
                if 'groupby' in constraints and constraints['groupby'] is not None:
                    groupby = constraints['groupby']
                    agg_func = agg_funcs[constraints['agg_func']]

                    if groupby in adata.obs:
                        grouped = adata.obs.groupby(groupby)[col].agg(agg_func)

                        group_gt = pd.Series([True] * len(grouped), index=grouped.index)
                        group_lt = pd.Series([True] * len(grouped), index=grouped.index)

                        if constraints['gt'] is not None:
                            group_gt = grouped >= constraints['gt']

                        if constraints['lt'] is not None:
                            group_lt = grouped <= constraints['lt']

                        group_keeper = group_gt & group_lt
                        groups_to_keep = group_keeper[group_keeper].index
                        cur_sub = adata.obs[groupby].isin(groups_to_keep)
                        col_keeper &= cur_sub
                else:
                    if constraints['gt'] is not None:
                        gt = column_data >= constraints['gt']

                    if constraints['lt'] is not None:
                        lt = column_data <= constraints['lt']

                    if constraints['exclude']:
                        exclude = ~column_data.isin(constraints['exclude'])

                    if constraints['subset'].items():
                        for subset_col, subset_values in constraints['subset'].items():
                            subset_in = adata.obs[subset_col].isin(subset_values) if subset_col in adata.obs else pd.Series([True] * adata.n_obs, index=adata.obs.index)
                            cur_sub = (gt & lt & exclude)
                            cur_sub |= ~subset_in
                            col_keeper &= cur_sub
                    else:
                        cur_sub = (gt & lt & exclude)
                        col_keeper &= cur_sub
                adata.uns['qc_filtered'][col][constraint_key] = list(adata.obs.index[~cur_sub])
                keeper_cells &= col_keeper


        for key in keys_to_remove:
            del adata.uns['qc_constraints'][key]
        
        adata.obs['keeper_cells'] = keeper_cells

        if inplace:
            adata._inplace_subset_obs(adata.obs['keeper_cells'])
            # Move the current qc_constraints to a uniquely named key and clear qc_constraints to fix weird bug hopefully
            count = 1
            new_key = f"qc_constraints_{count}"
            while new_key in adata.uns:
                count += 1
                new_key = f"qc_constraints_{count}"
            adata.uns[new_key] = adata.uns["qc_constraints"]
            adata.uns["qc_constraints"] = {}
        
        return adata



import seaborn as sns
import matplotlib.pyplot as plt
from typing import Sequence, Literal
from anndata import AnnData

def violin(
        adata: AnnData,
        y: str,
        groupby: str|None = None,
        log: bool = True,
        scale: Literal["area", "count", "width"] = "width",
        order: Sequence[str] | None = None,
        xlabel: str = "",
        size: int = 4,
        x_label_rotation: float = 0,
        ylines: Sequence[float] = []) -> tuple:
    """
    Compute entropy (mixing) of annotations within a cells local neighborhood. 

    Args:
        adata: Annotated data matrix.
        y: columns in `.obs` to plot for y axis.
        groupby: The key of the observation grouping to use for x axis.
        log: Use log scale for y.
        scale: The method used to scale the width of each violin.
        order: Order of x.
        xlabel: Label of the x axis. Defaults to `groupby`.
        size: Size of the x labels.
        x_label_rotation: Rotation of xtick labels.
        ylines: Add horizontal lines to the plot.
        
    Returns:
        Tuple[Axes, Axes]
    """

    count_values = adata.obs[groupby].value_counts().reset_index()
    count_values.columns = [groupby, 'cell_counts']
    if order is None:
        order = count_values[groupby].tolist()

    palette = sns.color_palette("husl", len(count_values))

    fig = plt.figure(figsize=(12, 8))
    gs = fig.add_gridspec(2, 1, height_ratios=[0.5, 2])  # Adjust ratios as needed


    # Create the bar plot on top of the violin plot
    ax1 = fig.add_subplot(gs[0, 0])
    sns.barplot(x=groupby, y='cell_counts', data=count_values, palette=palette, ax=ax1, order=order)
    ax1.set_xlabel('')  # Remove x-axis label
    ax1.set_xticks([])  # Remove x-ticks
    ax1.set_xticklabels([])  # Remove x-tick labels
    if log:
        ax1.set_yscale('log')

    # Create the violin plot
    ax2 = fig.add_subplot(gs[1, 0])
    sns.violinplot(x=groupby, y=y, data=adata.obs, inner=None, palette=palette, ax=ax2, order=order, density_norm=scale)
    ax2.set_xticklabels(ax2.get_xticklabels(), rotation=x_label_rotation, ha='right', size=size)  # Rotate x labels
    if log:
        ax2.set_yscale('log')
    if xlabel !="":
        ax2.set_xlabel(xlabel)
    for y in ylines:  # add horizontal lines
        ax2.axhline(y=y, color='red', linestyle='--', linewidth=1)

    plt.tight_layout()

    return ax1, ax2



# """
#     Violin plot.

#     Args:
#         adata: Annotated data matrix.
#         y: columns in `.obs` to plot for y axis.
#         groupby: The key of the observation grouping to use for x axis.
#         log: Use log scale for y.
#         scale: The method used to scale the width of each violin.
#             If 'width' (the default), each violin will have the same width.
#             If 'area', each violin will have the same area.
#             If 'count', a violin’s width corresponds to the number of observations.
#         order: Order of x.
#         xlabel: Label of the x axis. Defaults to `groupby`.
#         size: Size of the x labels.
#         x_label_rotation: Rotation of xtick labels.
#         ylines: Add horizontal lines to the plot.
        
#     Returns:
#         Tuple[Axes, Axes]
#         A tuple containing two matplotlib Axes objects:
#             The first Axes object is the bar plot showing the count of observations for each group.
#             The second Axes object is the violin plot showing the distribution of values in `y` for each group.
#     """


"""
sciduck

single cell data analyis unification and communication toolkit
"""

__version__ = "0.4.0"
__authors__ = ['Nelson Johansen', 'Matthew Schmitz', 'Dan Yuan']
__credits__ = 'Allen Institute for Brain Science'

from . import basic_qc as qc
from . import plotting as pl
from . import annotation_table as anno
from . import entropy_qc as eq
from . import templates as template



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
import pathlib
sys.path.insert(0, os.path.abspath('../../src/'))

# -- General configuration ------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
]

# Generate the API documentation when building
# autosummary_generate = True
# autodoc_member_order = 'bysource'

napoleon_google_docstring = True
napoleon_numpy_docstring = False

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'sciduck'
copyright = '2024, Nelson Johansen'
author = 'Nelson Johansen'

# Autodoc settings
autodoc_default_options = {
    'members': True,
    'undoc-members': False,  # Hide undocumented members
    'private-members': False,  # Hide private members (_member)
    'special-members': False,  # Hide special members (__member__)
    'inherited-members': False,  # Hide inherited members
    'show-inheritance': True,
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output
# pip install sphinx-rtd-theme; https://sphinx-themes.org/sample-sites/sphinx-rtd-theme/

# html_theme = 'sphinx_rtd_theme'



import numpy as np
import pandas as pd
from anndata import AnnData
import warnings

##
def allen_institute_basic(adata: AnnData):
    """
    Adds Allen Institute basic QC standards to the AnnData object. Does not perform the filtering.

    Args:
        adata: The AnnData object where constraints will be added.
    """
    ## Filter cells/nuclei on UMI and gene count thresholds, showing the default values.
    add_range_constraint(adata, column="counts", gt=2000, lt=100000)
    add_range_constraint(adata, column="genes", gt=1000, lt=13000)

    ## Filter cells/nuclei on mitochondrial gene expression, showing the default values.
    add_range_constraint(adata, "doublet_score", lt = 0.3)
    add_range_constraint(adata, "pct_counts_mt", lt = 3.0)
    add_range_constraint(adata, "GEX_Reads_mapped_confidently_to_genome", gt = 0.0)
    add_range_constraint(adata, "GEX_Reads_mapped_to_genome", gt = 0.0)
    add_range_constraint(adata, "GEX_Reads_with_TSO", lt = 1.0)

    ## Neuron / Non-Neuron QC constraints
    add_range_constraint(adata, "genes", gt = 2000, subset = "Class", subset_values = ['Excitatory', 'Inhibitory'])
    add_range_constraint(adata, "genes", gt = 1000, subset = "Class", subset_values = ['Astrocytes', 'Oligodendrocytes', 'Microglia', 'Endothelial', 'Pericytes'])

    return adata



from anndata import AnnData
import scanpy as sc
from scipy.spatial import KDTree
from scipy.stats import entropy
import numpy as np

def cell_entropy(adata: AnnData,
                    annotation_columns: list,
                    nearest_neighbors: int = 15, 
                    dim: str = "X_scVI"
                ) -> AnnData:
    """
    Compute entropy (mixing) of annotations within a cells local neighborhood. 

    Args:
        adata: AnnData object with `annotations`
        annotation_columns: Cell level annotations.
        nearest_neighbors: Number of nearest neighbors.
        dim: Dimensionality reduction
        
    Returns:
        Returns the updated AnnData object.
    """
    ## Build nearest neighbor tree for fast lookup
    print("Building nearest neighbor tree.")
    nnTree = KDTree(adata.obsm[dim])
    nearest_dist, nearest_ind = nnTree.query(adata.obsm[dim], k=nearest_neighbors)
    ##
    for anno in annotation_columns:
        print("Computing entropy on: " + anno)
        adata.obs[anno + "_entropy"] = -1 ## Initialize with a value outside range of entropy.
        for cell in range(0, adata.shape[0]):
            nearest_neighbors = nearest_ind[cell,:]
            adata.obs.loc[adata.obs.index[cell], anno + "_entropy"] = entropy(adata.obs.loc[adata.obs.index[nearest_neighbors],anno].value_counts()/len(nearest_neighbors))
    return adata

def cluster_entropy(adata: AnnData,
                    group_by: str,
                    annotation_columns: list
                ) -> AnnData:
    """
    Compute entropy (mixing) of annotations within a cell set (group_by). 

    Args:
        adata: AnnData object with `annotations`
        group_by: AnnData.obs column that groups cells.
        annotation_columns: Cell level annotations.
        
    Returns:
        Returns the updated AnnData object.
    """
    ##
    for anno in annotation_columns:
        print("Computing entropy on: " + anno)
        adata.obs[anno + "_entropy"] = adata.obs.groupby(group_by)[anno].value_counts(normalize=True).groupby(group_by).apply(lambda x: entropy(x))
    return adata
